import copy
import math
import sys
import getopt
from itertools import chain, combinations
from PyTASTileSystem import *
from square_gen import gen_square_tile_set, create_dummy_seed
from iu_gen import gen_iu_ttypes


def DisplayHelp():
    print("\nThis script generates a tile set used to create a ``quine'' from a given encoding of a partial glue lookup table.")
    print("\tThe created tile set computes the quine and completes the glue lookup table, then creates a macrotile for an IU simulation.")
    print("\nUsage:")
    print("-h, --help\t\t\tDisplays this help message.")
    print("-o <name>, --out=<name>\t\tThe base name of the TDP and TDS files to be output")
    print("-d <dir>, --dir=<dir>\t\t[Optional] Directory for output files. If not specified, the")
    print("\t\t\t\t current directory will be used.")
    print("-i <name>, --in=<name>\t\t[Optional] The name of the TDS file from which to read the IU tile set.")
    print("\t\t\t\t If neither this nor the 'iu' option are used, the IU tile set will be omitted.")
    print("-u <z,e>, --iu=<z,e>\t\t[Optional] If this option is used, the IU tile set will be generated")
    print("\t\t\t\t for fractal dimention z and epsilon e. If neither this nor the 'in' option are")
    print("\t\t\t\t used, the IU tile set will be omitted.")
    print("-s, --seed\t\t\t[Optional] If this option is used, the tiles for the seed row of")
    print("\t\t\t\t the quine will be included in the ouptut. Otherwise, they won't.")
    print("-t, --test\t\t\t[Optional] If this option is used, the tiles for the seed row of")
    print("\t\t\t\t the quine will be generated from a short 'test' string. Otherwise,")
    print("\t\t\t\t they'll be generated from the input tile set.")
    print("-r, --rect\t\t\t[Optional] If this option is used, the tiles that cause the quine rectangle")
    print("\t\t\t\t to grow into a square macrotile will be omitted.")
    print("-m <height>, --macro=<height>   [Optional] This option is for testing macrotile square generation only.")
    print("\t\t\t\t If this option is used, quine tiles are not created and an assembly to be used for")
    print("\t\t\t\t testing square macrotile creation will be generated with the specificed height.")
    exit()


"""
TODO: 
- After rotating for east side, will need to zero out g_E on the north and g_N on the east (probably)

Note: counter is flipped in relation to glues, with LSB on left, MSB on right for easier manipulation


This script generates the tiles used to create a ``quine'' as well as a macrotile representing the seed tile of the row
that represents the quine. The three major components are:
  1. The portion that takes as input a tile set and which creates the encoding of those tiles which should be presented
     as the northern glues of the seed row, encoding those tiles for the glue lookup table
  2. The portion that takes a list of symbols that can be exposed by the glue lookup table and generates the tile types
     that read the northern glues of the seed row, infer the full definitions of each tile, and adds those definitions
     to the right side of the glue lookup table encoded in the seed row.
  3. The portion that generates the tiles that continue growth of the assembly from the ocmpleted quine (i.e. full glue
     lookup table) into a full macrotile that encodes the seed tile on its sides.

  
TODO: NOTE: the ctr_start value must be padded enough to allow necessary increments!!!
    
    Formats and conventions

        Glue lookup table:
            T<tile1 input>#<tile1 output>#<tile2 input>#<tile2 output>#...#<tileN input>#<tileN output>T

        Tile definition:
                Input glues    Output glues
            <gN><gE><gS><gW>#<gN><gE><gS><gW>
            
                - each gD consists of either a 'w' or an 's', and may be followed by a binary number
                    - 'w': weak, i.e. strength-1 glue
                    - 's': strong, i.e. strength-2 glue
                    - if the glue is null, 'w' is used and no binary number follows
                    - if the glue is not null, the binary number associated with the glue label follows
        
        'T' marks the left side and right sides of the lookup table
        '#' in between tile definitions and in between the input and output glue definitions of each tile

        Example glue lookup table (representing a single tile): Twws10010w#w1010w1111T

        Seed row:
            T<ctr_start><glue lookup table>
                - ctr_start: the value to be used for the first E glue of a seed row tile type to be added to the glue table (with the right end marked by 'Z' instead of 'T')
                - glue lookup table: glue lookup table (excluding seed row tiles)

        Example seed row: #11000Twws10010w#w1010w1111Z
            - ctr_start: 11000
            - glue lookup table: Twws10010w#w1010w1111Z

        Glue symbol mapping: <null>:0, 1:1, 0:10, w:11, s:100, #:101, T:110

        Example tile encodings:
          leftmost # tile of seedrow (north glue strength-1 'T', east glue strength-2 '<ctr_start>':  wwww#w101s<ctr_start>ww
          leftmost 1 tile of seedrow (north glue strength-1 '1', east glue strength-2 '<ctr_start+1>', west glue strength-2 '<ctr_start>':  wwws<ctr_start>#w1s<ctr_start+1>ww
          

    Growth phases:
        1. Grow seed row
            - Grows left to right

        2. Add definitions of seed row tiles at right end of glue lookup table (skipping 'counter' glue)
            - Even rows: grow right to left, looking for final # and copying S to N
            - Odd rows: grow left to right, remember first umarked symbol and mark it, put 'wwwsc#w1scww' before R (where 'c' will be replaced with current counter value)

        3. Add counter glues

        4. Expand glue table using padding

        5. Turn glue representation length and glue counter start value sections into spacers

        6. Add definitions of seed tile output glues to top row

"""

##############################################################################
# Constants for directions, etc.
N = North = 0
E = East = 1
S = South = 2
W = West = 3
U = Up = 4
D = Down = 5

Dirs = [N,E,S,W]
DirAbbrevArray = ("N", "E", "S", "W", "U", "D")
DirNameArray = ("NORTH", "EAST", "SOUTH", "WEST", "UP", "DOWN")

#INPUT_MARKERS = {N: 'v', E: '<', S: '^', W: '>'}
#OUTPUT_MARKERS = {N: '^', E: '>', S: 'v', W: '<'}
INPUT_MARKERS = {N: 'vv', E: '<<', S: '^^', W: '>>'}
OUTPUT_MARKERS = {N: '^^', E: '>>', S: 'vv', W: '<<'}
SPECIAL_OUTPUT_MARKERS = {N: '^', E: '>', S: 'v', W: '<'}
IU_INPUT_MARKERS = {N: 'v', E: '<', S: '^', W: '>'}
IU_OUTPUT_MARKERS = {N: '^', E: '>', S: 'v', W: '<'}

OPP_DIR = {N: S, E: W, S: N, W: E}

color_list = ["white", "green", "red", "blue", "yellow", "pink", "aqua", "lime", "olive",
              "orange", "gold", "grey", "purple", "maroon", "turquoise", "black"]


# This is a helper function to convert from IU IO markers to the IO markers used by the quine tiles
def fix_IO_markings(squaring_ttypes : list) -> dict:
    squaring_ttypes_dict = {}
    for squaring_ttype in squaring_ttypes:
        tt_name = squaring_ttype.name
        for side in [N,E,S,W]:
            if (squaring_ttype.sides[side] is not None) and (squaring_ttype.sides[side].label.endswith(IU_INPUT_MARKERS[side])) and (not squaring_ttype.sides[side].label.endswith(INPUT_MARKERS[side])):
                squaring_ttype.sides[side].label += IU_INPUT_MARKERS[side]
            elif (squaring_ttype.sides[side] is not None) and (squaring_ttype.sides[side].label.endswith(IU_OUTPUT_MARKERS[side])) and (not squaring_ttype.sides[side].label.endswith(OUTPUT_MARKERS[side])):
                squaring_ttype.sides[side].label += IU_OUTPUT_MARKERS[side]
        squaring_ttypes_dict[tt_name] = squaring_ttype
    return squaring_ttypes_dict


def draw_tile_set(tile_set, output_directory, base_file_name, b_output):
    if not b_output:
        return
    system = PyTASTileSystem()
    system.tileTypes = tile_set
    system.DrawTileSet(pdf_file_name = f'{output_directory}/{base_file_name}.pdf', svg_file_name = f'{output_directory}/{base_file_name}.svg')

def get_glue_sets_for_dirs(tile_set_dict):
    glue_sets = {N: set(), E: set(), S: set(), W: set()}
    for tile_type_name in tile_set_dict:
        tile_type = tile_set_dict[tile_type_name]
        for d in Dirs:
            side = tile_type.sides[d]
            if (not side is None) and (side.strength > 0):
                #glue_sets[d].add((side.strength, side.label))
                glue_sets[d].add(side.label)
    return glue_sets

def get_horizontal_glue_set(tile_set_dict):
    glue_sets = get_glue_sets_for_dirs(tile_set_dict)
    return glue_sets[E].union(glue_sets[W])

def get_vertical_glue_set(tile_set_dict):
    glue_sets = get_glue_sets_for_dirs(tile_set_dict)
    return glue_sets[N].union(glue_sets[S])

# B0-B7 : field delimiters, in increasing order from left to right
# $ : start of tile encoding
# |1 : 1 of 2 characters between input and output glue encodings of a tile
# |2 : 2 of 2 characters between input and output glue encodings of a tile
# # : start of glue encoding
# n : null glue
# p : pad
seed_row_symbol_list      = ['B0!', 'B1', 'B2', 'B3', 'B4', '$', '|1', '|2', '#', 'n', 'p', '0', '1']
post_seed_row_symbol_list = ['B0',  'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', '$', '|1', '|2', '#', 'n', 'p', '0', '1']

def get_seed_tile_template_tiles(curr_phase_marker, num_bits_for_symbols, color, b_compress=True):
    tile_set = dict()
    for seed_tile_symbol in seed_row_symbol_list:
        symbol_index = seed_row_symbol_list.index(seed_tile_symbol) + 1
        binary_symbol_list = list(bin(symbol_index)[2:])
        if num_bits_for_symbols < len(binary_symbol_list):
            print(f'ERROR: Not enough bits to encoded the seed row tiles!')
            quit()
        binary_symbol_list = ['0'] * (num_bits_for_symbols - len(binary_symbol_list)) + binary_symbol_list
        if seed_tile_symbol == 'B0!':  # No input west glue, strength-2 output north glue, strength-2 output east glue (ctr glue will have 1-bit for strength-2 added automatically later)
            template_list = ['#', 'n', '#', 'n', '#', 'n', '#', 'n', '|1', '|2', '#', '1', 'p'] + binary_symbol_list + ['#', 'ctr', '#', 'n', '#', 'n']
        elif seed_tile_symbol == 'B4': # Strength-2 input west glue, strength-1 output north glue, no output east glue
            template_list = ['#', 'n', '#', 'n', '#', 'n', '#', 'ctr', '|1', '|2', '#', '0', 'p'] + binary_symbol_list + ['#', 'n', '#', 'n', '#', 'n']
        else: # Strength-2 input west glue, strength-1 output north glue, strength-2 output east glue
            template_list = ['#', 'n', '#', 'n', '#', 'n', '#', 'ctr', '|1', '|2', '#', '0', 'p'] + binary_symbol_list + ['#', 'ctr', '#', 'n', '#', 'n']

        if not b_compress:
            new_template_list = []
            for sym in template_list:
                if sym == 'n':
                    new_template_list += ['0'] * (num_bits_for_symbols + 1) # Include the extra 0 for the strength bit
                elif sym != 'p':
                    new_template_list.append(sym)
            template_list = new_template_list

        template_list.append('B4')
        #tl_str = ''
        #print(f'Template list len for {seed_tile_symbol} with num_bits_for_symbols = {num_bits_for_symbols}  is {len(template_list)}: {tl_str.join(template_list)}')
        count = 0
        for symbol in template_list:
            tile_type = PyTASTileType()
            tile_type.name = f'Ext_{seed_tile_symbol}_{count}'
            tile_type.label = tile_type.name
            tile_type.color = color

            if count == 0:
                tile_type.sides[W] = Glue(2, f'{curr_phase_marker}_{seed_tile_symbol}')
            else:
                tile_type.sides[W] = Glue(2, f'{tile_type.name}')

            if count < (len(template_list) - 1):
                tile_type.sides[E] = Glue(2, f'Ext_{seed_tile_symbol}_{count+1}')
                tile_type.sides[N] = Glue(1, f'{symbol}')
            else:
                tile_type.sides[N] = Glue(2, f'{symbol}')
                tile_type.sides[E] = Glue(1, 'FILL')

            # Special case for first extension tile for first seed row tile (B0!) to have a south glue that allows filler to attach
            if (seed_tile_symbol == 'B0!') and (count == 0):
                tile_type.sides[S] = Glue(2, 'Ext_FILL')
            else:
                tile_type.sides[S] = Glue(1, 'FILL')
            
            IO_mark_tile_type(tile_type, [W])
            tile_set[tile_type.name] = tile_type
            count += 1

    return tile_set

#seed_row_symbol_list_len = len(seed_row_symbol_list)
#counter_start_val = seed_row_symbol_list_len
#seed_row_symbol_binary_encoding_len = bin(seed_row_symbol_list_len-1)[2:]

def get_padded_symbol_val_list(glue_symbol, b_horiz, horizontal_symbol_list, vertical_symbol_list):
    list_len = len(horizontal_symbol_list) if len(horizontal_symbol_list) > len(vertical_symbol_list) else len(vertical_symbol_list)
    if b_horiz:
        index = horizontal_symbol_list.index(glue_symbol)
    else:
        index = vertical_symbol_list.index(glue_symbol)
    binary_list = list(bin(index)[2:])
    binary_list = ['0'] * (len(bin(list_len)[2:]) - len(binary_list)) + binary_list
    return binary_list

# def get_list_with_padding_from_binary_string(binary_string):
#     ret_list = list(binary_string)
#     ret_list = ret_list[:1] + ['p'] + ret_list[1:]
#     return ret_list

def get_glue_encoding_list(tile_type, padded_len, horizontal_symbol_list, vertical_symbol_list, b_input=True, iu_set_markings=False):
    glue_encoding_list = []
    b_found = False
    for d in Dirs:
        if iu_set_markings:
            io_marker = IU_INPUT_MARKERS[d] if b_input else IU_OUTPUT_MARKERS[d]
        else:
            io_marker = INPUT_MARKERS[d] if b_input else OUTPUT_MARKERS[d]
        
        if (not tile_type.sides[d] is None) and (tile_type.sides[d].strength > 0) and tile_type.sides[d].label.endswith(io_marker):
            if d in [N,S]:
                encoding_val = vertical_symbol_list.index(tile_type.sides[d].label)
            else:
                encoding_val = horizontal_symbol_list.index(tile_type.sides[d].label)
            strength_bit = '0' if tile_type.sides[d].strength == 1 else '1'
            binary_encoding_list = list(bin(encoding_val)[2:])
            glue_encoding_list += ['#', strength_bit, 'p'] + ['0']*(padded_len - len(binary_encoding_list)) + binary_encoding_list
            b_found = True
        else:
            glue_encoding_list += ['#', 'n']
    if not b_found:
        if b_input:
            print(f'WARNING: No input sides found for tile type {tile_type.name} with iu_set_markings = {iu_set_markings}')
        else:
            print(f'WARNING: No output sides found for tile type {tile_type.name} with iu_set_markings = {iu_set_markings}')
    return glue_encoding_list

def get_input_glue_encoding_list(tile_type, padded_len, horizontal_symbol_list, vertical_symbol_list, iu_set_markings=False):
    return get_glue_encoding_list(tile_type, padded_len, horizontal_symbol_list, vertical_symbol_list, True, iu_set_markings=iu_set_markings)

def get_output_glue_encoding_list(tile_type, padded_len, horizontal_symbol_list, vertical_symbol_list, iu_set_markings=False):
    return get_glue_encoding_list(tile_type, padded_len, horizontal_symbol_list, vertical_symbol_list, False, iu_set_markings=iu_set_markings)

def get_T_list_from_tile_set(tile_set, max_glue_value_len, horizontal_symbol_list, vertical_symbol_list, iu_set_markings):
    T_list = []
    for tile_type_name in tile_set:
        tile_type = tile_set[tile_type_name]
        tile_list = ['$']
        tile_list += get_input_glue_encoding_list(tile_type, max_glue_value_len, horizontal_symbol_list, vertical_symbol_list, iu_set_markings)
        tile_list += ['|1', '|2']
        tile_list += get_output_glue_encoding_list(tile_type, max_glue_value_len, horizontal_symbol_list, vertical_symbol_list, iu_set_markings)
        #print(f'tile_list: {tile_list}')
        T_list += tile_list
    return T_list

def get_seed_row_encoding_list(g_N, quine_tile_set, iu_tile_set, starting_counter_width, b_show_warnings, b_testing=False):
    if quine_tile_set is None:
        combined_tile_set = dict()
    else:
        combined_tile_set = copy.deepcopy(quine_tile_set)

    if not iu_tile_set is None:
        #print(f'DEBUG: Have IU tile set of size {len(iu_tile_set)}')
        combined_tile_set.update(iu_tile_set)
    #else:
        #print(f'DEBUG: No IU tile set')
    
    horizontal_glue_set = get_horizontal_glue_set(combined_tile_set)
    vertical_glue_set = get_vertical_glue_set(combined_tile_set)
    for symbol in seed_row_symbol_list:
        if not symbol in vertical_glue_set:
            if b_show_warnings:
                print(f'WARNING: symbol {symbol} from the north side of the seed row is not found as a southern glue of any tile in the tile set')
            vertical_glue_set.add(symbol)
    
    horizontal_symbol_list = sorted(list(horizontal_glue_set))
    vertical_symbol_list = sorted(list(vertical_glue_set))

    # Offset to reserve encoding of 0 for the null glue
    horizontal_symbol_list = [''] + horizontal_symbol_list
    vertical_symbol_list = [''] + vertical_symbol_list

    max_glue_set_len = max(len(horizontal_glue_set), len(vertical_glue_set))
    max_glue_value = max_glue_set_len
    max_glue_value_len = len(bin(max_glue_value)[2:])

    T_list = []
    if not quine_tile_set is None:
        quine_T_list = get_T_list_from_tile_set(quine_tile_set, max_glue_value_len, horizontal_symbol_list, vertical_symbol_list, False)
        T_list += quine_T_list
    if not iu_tile_set is None:
        iu_T_list = get_T_list_from_tile_set(iu_tile_set, max_glue_value_len, horizontal_symbol_list, vertical_symbol_list, True)
        T_list += iu_T_list
    #print(f'\nT_list: {T_list}\n')

    g_N_list = get_padded_symbol_val_list(g_N, False, horizontal_symbol_list, vertical_symbol_list)

    longest_list_len = len(horizontal_symbol_list) if len(horizontal_symbol_list) > len(vertical_symbol_list) else len(vertical_symbol_list)
    counter_start_bin_val = bin(longest_list_len)[2:]
    counter_start_bin_val_list = list(counter_start_bin_val)
    padding_list = ['0']*(starting_counter_width - len(counter_start_bin_val_list))
    # if counter_start_val is exactly a power of 2, will need to add another bit to the padding
    #  since encoded glues will be at least one value less and thus one bit shorter than the initial counter value
    if (longest_list_len & (longest_list_len-1) == 0) and longest_list_len != 0:
        padding_list.append('0')
    padded_ctr_start_val_list = padding_list + counter_start_bin_val_list

    # FIX: Must use uncompressed value of g_N since expansion of p is done only for values to its right
    g_N_list = ['0'] * (starting_counter_width - len(g_N_list)) + g_N_list

    # g_E will be the starting value of the counter, since that will be the value used to encode the eastern glue of the seed tile and it's the first ctr value used
    g_E_list = padded_ctr_start_val_list

    if b_testing:
        g_N_list = ['0', '0']
        g_E_list = ['0', '0']

    # Flipping the bits of the counter so they can be more easily utilized
    reversed_padded_ctr_start_val_list = padded_ctr_start_val_list[::-1]

    seed_row_list =  ['B0!'] + reversed_padded_ctr_start_val_list + ['B1', '#', '1'] + g_N_list + ['#', '1'] + g_E_list + ['#', 'n', '#', 'n', 'B2'] + padding_list + ['B3'] + T_list + ['B4']

    return seed_row_list, longest_list_len

##################################################################################################
####    Functions to build sets of tiles for various functions


# Build the set of tiles to assemble the 'seed' row encoding the list of symbols given
def get_seed_row_tile_set(seed_row_list):
    seed_row_tile_set = dict()
    seed_row_tile_list = []
    seed_tile_type_name = None
    counter = 0
    for seed_row_value in seed_row_list:

        tile_type = PyTASTileType()
        tile_type.name = f'SeedRow{counter}'
        tile_type.label = tile_type.name
        tile_type.color = 'green'
        input_sides = []

        if counter == 0:
            seed_tile_type_name = tile_type.name
            tile_type.sides[N] = Glue(2, f'{seed_row_value}')
        else:
            tile_type.sides[N] = Glue(1, f'{seed_row_value}')
            tile_type.sides[W] = Glue(2, f'S{counter}')
            input_sides.append(W)
        if counter < (len(seed_row_list) - 1):
            tile_type.sides[E] = Glue(2, f'S{counter+1}')
        counter += 1

        IO_mark_tile_type(tile_type, input_sides)
        seed_row_tile_set[tile_type.name] = tile_type
        seed_row_tile_list.append(tile_type)

    return (seed_tile_type_name, seed_row_tile_set, seed_row_tile_list)

# Build the set of tiles that take a value from the west, pass it north, and pass their south value east
def get_right_shift_tiles(hor_prefix, vert_symbol_list, right_end_symbol, color):
    right_shift_tile_set = dict()

    # Make tile types for row interior positions
    for west_symbol in vert_symbol_list:
        for south_symbol in vert_symbol_list:

            tile_type = PyTASTileType()
            tile_type.name = f'Shift_{hor_prefix}_{west_symbol}_{south_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[N] = Glue(1, f'{west_symbol}')
            if south_symbol == right_end_symbol:
                tile_type.sides[E] = Glue(2, f'{hor_prefix}_{south_symbol}')
            else:
                tile_type.sides[E] = Glue(1, f'{hor_prefix}_{south_symbol}')
            tile_type.sides[S] = Glue(1, f'{south_symbol}')
            tile_type.sides[W] = Glue(1, f'{hor_prefix}_{west_symbol}')

            IO_mark_tile_type(tile_type, [S,W])

            right_shift_tile_set[tile_type.name] = tile_type

    # Make right extension tile type
    tile_type = PyTASTileType()
    tile_type.name = f'Shift_{hor_prefix}_{right_end_symbol}'
    tile_type.label = tile_type.name
    tile_type.color = color
    tile_type.sides[N] = Glue(2, f'{hor_prefix}_{right_end_symbol}')
    tile_type.sides[W] = Glue(2, f'{hor_prefix}_{right_end_symbol}')
    tile_type.sides[S] = Glue(1, f'FILL')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [W])
    right_shift_tile_set[tile_type.name] = tile_type

    return right_shift_tile_set

def get_tiles_to_pass_symbols_right(prefix, horiz_symbol_list, vert_symbol_list, vert_symbol_end, vert_symbol_new_val, color):
    tile_set = dict()
    for horiz_symbol in horiz_symbol_list:
        for vert_symbol in vert_symbol_list:
            tile_type = PyTASTileType()
            tile_type.name = f'PassRight_{prefix}_{horiz_symbol}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[W] = Glue(1, f'{prefix}_{horiz_symbol}')
            if vert_symbol == vert_symbol_end:
                tile_type.sides[N] = Glue(1, f'{vert_symbol_new_val}')
                tile_type.sides[E] = Glue(2, f'{prefix}_{horiz_symbol}')
            else:
                tile_type.sides[N] = Glue(1, f'{vert_symbol}')
                tile_type.sides[E] = Glue(1, f'{prefix}_{horiz_symbol}')

            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type

            # Make tiles that receive 'pass' from left
            tile_type = PyTASTileType()
            tile_type.name = f'PassRight_pass_{prefix}_{horiz_symbol}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[W] = Glue(1, f'{prefix}_pass_{horiz_symbol}')
            if vert_symbol == vert_symbol_end:
                tile_type.sides[N] = Glue(1, f'{prefix}_{vert_symbol_new_val}')
                tile_type.sides[E] = Glue(2, f'{prefix}_{horiz_symbol}')
            else:
                tile_type.sides[N] = Glue(1, f'{prefix}_{vert_symbol}')
                tile_type.sides[E] = Glue(1, f'{prefix}_{horiz_symbol}')

            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type

    return tile_set

W_to_E = 0
E_to_W = 1

def get_tiles_to_find_symbol(horiz_symbol_before, found_prefix, vert_symbol_target, vert_symbol_list, color, direction, swap_symbol=None):
    symbol_finding_tiles = dict()
    in_dir = W if direction == W_to_E else E
    out_dir = E if direction == W_to_E else W
    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'Find{vert_symbol_target}_{horiz_symbol_before}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[in_dir] = Glue(1, f'{horiz_symbol_before}')
        if vert_symbol == vert_symbol_target:
            if swap_symbol is not None:
                tile_type.sides[N] = Glue(1, f'{swap_symbol}')
            else:
                tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[out_dir] = Glue(1, f'{found_prefix}')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[out_dir] = Glue(1, f'{horiz_symbol_before}')

        if direction == W_to_E:
            IO_mark_tile_type(tile_type, [S,W])
        else:
            IO_mark_tile_type(tile_type, [S,E])
        symbol_finding_tiles[tile_type.name] = tile_type
    return symbol_finding_tiles

def get_tiles_to_get_symbol_at_marker(marker, vert_symbol_list, color, direction):

    if marker in vert_symbol_list:
        print(f'ERROR: in get_tiles_to_get_symbol_at_marker marker {marker} should not be in  vert_symbol_list {vert_symbol_list}')
        quit()

    in_dir = W if direction == W_to_E else E
    out_dir = E if direction == W_to_E else W

    tile_set = dict()
    for vert_symbol in vert_symbol_list:
        # Make tile for before marker is found
        tile_type = PyTASTileType()
        tile_type.name = f'Find{marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[in_dir] = Glue(1, f'{marker}')
        tile_type.sides[out_dir] = Glue(1, f'{marker}')

        if direction == W_to_E:
            IO_mark_tile_type(tile_type, [S,W])
        else:
            IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type


        # Make tile for tile with marker
        marked_vert_symbol = f'{marker}_{vert_symbol}'

        tile_type = PyTASTileType()
        tile_type.name = f'Found{marked_vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        tile_type.sides[S] = Glue(1, f'{marked_vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[in_dir] = Glue(1, f'{marker}')
        tile_type.sides[out_dir] = Glue(1,  f'{marker}_pass_{vert_symbol}')

        if direction == W_to_E:
            IO_mark_tile_type(tile_type, [S,W])
        else:
            IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type

    return tile_set

def get_tiles_for_left_reset(vert_symbol_list, right_end_symbol, prefix, right_end_prefix, color):
    tile_set = dict()
    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RtoL_R_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        if vert_symbol == right_end_symbol:
            if (right_end_prefix is not None) and (len(right_end_prefix) > 0):
                tile_type.sides[S] = Glue(2, f'{right_end_prefix}_{vert_symbol}')
            else:
                tile_type.sides[S] = Glue(2, f'{vert_symbol}')
            input_sides = [S]
            tile_type.sides[E] = Glue(1, f'FILL')
        else:
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'RtoL_{prefix}')
            input_sides = [S,E]
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'RtoL_{prefix}')
        IO_mark_tile_type(tile_type, input_sides)
        tile_set[tile_type.name] = tile_type
    return tile_set
    
def get_tiles_for_left_reset_with_prefix_check(vert_symbol_list, right_end_symbol, prefix, right_end_prefix, color):
    tile_set = dict()
    for vert_symbol in vert_symbol_list:
        # Make tiles for before encountering marker
        tile_type = PyTASTileType()
        tile_type.name = f'RtoL_R_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        if vert_symbol == right_end_symbol:
            if (right_end_prefix is not None) and (len(right_end_prefix) > 0):
                tile_type.sides[S] = Glue(2, f'{right_end_prefix}_{vert_symbol}')
            else:
                tile_type.sides[S] = Glue(2, f'{vert_symbol}')
            input_sides = [S]
            tile_type.sides[E] = Glue(1, f'FILL')
        else:
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'RtoL_{prefix}')
            input_sides = [S,E]
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'RtoL_{prefix}')
        IO_mark_tile_type(tile_type, input_sides)
        tile_set[tile_type.name] = tile_type

        # Make tiles for location of marker
        tile_type = PyTASTileType()
        tile_type.name = f'RtoL_at_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(1, f'{prefix}_{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{prefix}_{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'RtoL_{prefix}')
        tile_type.sides[W] = Glue(1, f'RtoL_ok_{prefix}')
        IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type

        # Make tiles for after encountering marker
        tile_type = PyTASTileType()
        tile_type.name = f'RtoL_L_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'RtoL_ok_{prefix}')
        tile_type.sides[W] = Glue(1, f'RtoL_ok_{prefix}')
        IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type
    return tile_set

def get_tiles_to_go_right(hor_prefix, vert_symbol_list, right_end_symbol, right_end_prefix, color):
    tile_set = dict()
    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'LtoR_{hor_prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        if vert_symbol == right_end_symbol:
            if (right_end_prefix is not None) and (len(right_end_prefix) > 0):
                tile_type.sides[N] = Glue(2, f'{right_end_prefix}_{vert_symbol}')
            else:
                tile_type.sides[N] = Glue(2, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'FILL')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{hor_prefix}')
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'{hor_prefix}')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type
    return tile_set

def get_tiles_to_copy_region_to_right(prefix, region_src_start_symbol, region_src_end_symbol, region_dst_start_symbol, region_dst_end_symbol,
                                      region_symbol_list, vert_symbol_list, color):
    tile_set = dict()

    # Mark src start and dst
    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY1_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        if vert_symbol == region_src_start_symbol:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}_SRC')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY1')
        elif vert_symbol == region_dst_start_symbol:
            tile_type.sides[N] = Glue(2, f'{prefix}_{vert_symbol}_DST')
            tile_type.sides[E] = Glue(1, f'FILL')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY1')
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY1')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type

    # Create vert_symbol_list for marked src and dst symbols
    src_marked_symbol_list = [f'{region_src_end_symbol}_SRC']
    if region_src_start_symbol != 'B0':
        src_marked_symbol_list.append(f'{region_src_start_symbol}_SRC')
    dst_marked_symbol_list = [f'{region_dst_start_symbol}_DST', f'{region_dst_end_symbol}_DST']
    for symbol in region_symbol_list:
        src_marked_symbol_list.append(f'{symbol}_SRC')
        dst_marked_symbol_list.append(f'{symbol}_DST')

    # Return left
    for vert_symbol in dst_marked_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY_LFT_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(2, f'{prefix}_{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY_LFT')
        tile_type.sides[E] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [S])
        tile_set[tile_type.name] = tile_type

    for vert_symbol in vert_symbol_list + src_marked_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY_LFT_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        if vert_symbol == 'B0':
            tile_type.sides[N] = Glue(2, f'{prefix}_B0_RGNCPY2')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY_LFT')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY_LFT')
        IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type

    # Go right to find marked src
    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        if vert_symbol == 'B0':
            tile_type.sides[S] = Glue(2, f'{prefix}_B0_RGNCPY2')
            input_sides = [S]
        else:
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            input_sides = [S,W]
        tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY2')
        IO_mark_tile_type(tile_type, input_sides)
        tile_set[tile_type.name] = tile_type

    if region_src_start_symbol == 'B0':
        # Handle special case where B0 is the src marked symbol
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY_LFT_{prefix}_B0*'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(1, f'B0_SRC')
        tile_type.sides[N] = Glue(2, f'{prefix}_B0_SRC_M')
        tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY_LFT')
        IO_mark_tile_type(tile_type, [S,E])
        tile_set[tile_type.name] = tile_type

        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY2_{prefix}_B0_M1'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(2, f'{prefix}_B0_SRC_M')
        tile_type.sides[N] = Glue(1, f'B0')
        tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2_M')
        IO_mark_tile_type(tile_type, [S])
        tile_set[tile_type.name] = tile_type

    # At marked src, move mark right and carry newly marked symbol to end
    for vert_symbol in src_marked_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}_M'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol[:-4]}')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY2')
        tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2_M')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type

    for vert_symbol in region_symbol_list + [region_src_end_symbol]:
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}_M2'
        tile_type.label = tile_type.name
        tile_type.color = color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        if vert_symbol == region_src_end_symbol:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2_{region_dst_end_symbol}')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}_SRC')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2_{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY2_M')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type

    for vert_symbol in vert_symbol_list:
        for passed_symbol in region_symbol_list + [region_dst_end_symbol]:
            tile_type = PyTASTileType()
            tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}_{passed_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY2_{passed_symbol}')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY2_{passed_symbol}')
            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type

    for passed_symbol in region_symbol_list + [region_dst_end_symbol]:
        for vert_symbol in dst_marked_symbol_list:
            tile_type = PyTASTileType()
            tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}_{passed_symbol}_R1'
            tile_type.label = tile_type.name
            tile_type.color = color
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{vert_symbol[:-4]}')
            tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY2_{passed_symbol}')
            tile_type.sides[E] = Glue(2, f'{prefix}_RGNCPY2_{passed_symbol}_R')
            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type
        
        tile_type = PyTASTileType()
        tile_type.name = f'RGNCPY2_{prefix}_{vert_symbol}_{passed_symbol}_R2'
        tile_type.label = tile_type.name
        tile_type.color = color
        if passed_symbol == region_dst_end_symbol:
            tile_type.sides[N] = Glue(2, f'{prefix}_{passed_symbol}_DONE')
        else:
            tile_type.sides[N] = Glue(2, f'{prefix}_{passed_symbol}_DST')
        tile_type.sides[W] = Glue(2, f'{prefix}_RGNCPY2_{passed_symbol}_R')
        tile_type.sides[E] = Glue(1, f'FILL')
        tile_type.sides[S] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [W])
        tile_set[tile_type.name] = tile_type

    # Final tile for right end
    tile_type = PyTASTileType()
    tile_type.name = f'RGNCPY_DONE_{prefix}_{region_dst_end_symbol}'
    tile_type.label = tile_type.name
    tile_type.color = color
    tile_type.sides[S] = Glue(2, f'{prefix}_{region_dst_end_symbol}_DONE')
    tile_type.sides[N] = Glue(1, f'{region_dst_end_symbol}')
    tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY_DONE')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S])
    tile_set[tile_type.name] = tile_type

    # Final row
    for vert_symbol in vert_symbol_list:
        if vert_symbol != 'B0':
            tile_type = PyTASTileType()
            tile_type.name = f'RGNCPY_DONE_{prefix}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[W] = Glue(1, f'{prefix}_RGNCPY_DONE')
            tile_type.sides[E] = Glue(1, f'{prefix}_RGNCPY_DONE')
            IO_mark_tile_type(tile_type, [S,E])
            tile_set[tile_type.name] = tile_type

    return tile_set


def get_cleanup_tiles(vert_symbol_list, swap_pairs, end_symbol, prefix, color, direction):
    tile_set = dict()
    final_tile = None

    in_dir = W if direction == W_to_E else E
    out_dir = E if direction == W_to_E else W

    for vert_symbol in vert_symbol_list:
        tile_type = PyTASTileType()
        tile_type.name = f'Cleanup_{prefix}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        S_N_symbols = (vert_symbol, vert_symbol)
        for swap_pair in swap_pairs:
            if swap_pair[0] == vert_symbol:
                S_N_symbols = swap_pair
        tile_type.sides[S] = Glue(1, f'{S_N_symbols[0]}')
        if vert_symbol == end_symbol:
            tile_type.sides[N] = Glue(2, f'{S_N_symbols[1]}')
            tile_type.sides[E] = Glue(1, f'FILL')
            final_tile = tile_type
        else:
            tile_type.sides[N] = Glue(1, f'{S_N_symbols[1]}')
            tile_type.sides[E] = Glue(1, f'Cleanup_{prefix}')
        tile_type.sides[W] = Glue(1, f'Cleanup_{prefix}')
        IO_mark_tile_type(tile_type, [S,in_dir])
        tile_set[tile_type.name] = tile_type

    return tile_set, final_tile

def mark_first_unmarked(prefix, marker, hor_before_marker, hor_after_marker, vert_symbols, end_symbol, color, b_append_value=False):
    tile_set = dict()
    for vert_symbol in vert_symbols:
        # Marked
        tile_type = PyTASTileType()
        tile_type.name = f'MarkFirstUnmarked_{prefix}_{hor_before_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        tile_type.sides[S] = Glue(1, f'{marker}_{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{marker}_{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'MFUS_{prefix}_{hor_before_marker}')
        tile_type.sides[W] = Glue(1, f'MFUS_{prefix}_{hor_before_marker}')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type

        # Unmarked
        tile_type = PyTASTileType()
        tile_type.name = f'MarkFirstUnmarked_{prefix}_{hor_before_marker}_{marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = color

        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{marker}_{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'{prefix}_{hor_after_marker}')
        if b_append_value:
            tile_type.sides[E].label += f'_{vert_symbol}'
        tile_type.sides[W] = Glue(1, f'MFUS_{prefix}_{hor_before_marker}')
        IO_mark_tile_type(tile_type, [S,W])
        tile_set[tile_type.name] = tile_type

    return tile_set


def get_first_unmarked_sym1_after_sym2_tiles(prefix, marker, before_marker, after_marker, vert_symbol_list1, vert_symbol_list2, symbol1, symbol2_list, color):
    tile_set = dict()

    # Get tiles for before arriving at symbol1
    for vert_symbol in vert_symbol_list1:
        if vert_symbol != symbol1:
            tile_type = PyTASTileType()
            tile_type.name = f'MarkS1AfterS2_{prefix}_{before_marker}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'MS1AS2_{prefix}_{before_marker}')
            tile_type.sides[W] = Glue(1, f'MS1AS2_{prefix}_{before_marker}')
            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type

    # Get the tile for arriving at symbol1
    tile_type = PyTASTileType()
    tile_type.name = f'MarkS1AfterS2_{prefix}_{before_marker}_{symbol1}'
    tile_type.label = tile_type.name
    tile_type.color = color

    tile_type.sides[S] = Glue(1, f'{symbol1}')
    tile_type.sides[N] = Glue(1, f'{symbol1}')
    tile_type.sides[E] = Glue(1, f'MS1AS2_{prefix}_{after_marker}')
    tile_type.sides[W] = Glue(1, f'MS1AS2_{prefix}_{before_marker}')
    IO_mark_tile_type(tile_type, [S,W])
    tile_set[tile_type.name] = tile_type

    # Get the tiles for after symbol1, looking for unmarked symbol2
    for vert_symbol in vert_symbol_list2:

        if vert_symbol in symbol2_list:

            # Already marked
            tile_type = PyTASTileType()
            tile_type.name = f'MarkS1AfterS2_{prefix}_{after_marker}_{marker}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[S] = Glue(1, f'{marker}_{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{marker}_{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'MS1AS2_{prefix}_{after_marker}')
            tile_type.sides[W] = Glue(1, f'MS1AS2_{prefix}_{after_marker}')
            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type

            # First unmarked
            tile_type = PyTASTileType()
            tile_type.name = f'MarkS1AfterS2_{prefix}_{after_marker}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = color

            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{marker}_{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{prefix}')
            tile_type.sides[W] = Glue(1, f'MS1AS2_{prefix}_{after_marker}')
            IO_mark_tile_type(tile_type, [S,W])
            tile_set[tile_type.name] = tile_type


    return tile_set

def get_symbol_passing_tiles(tile_type, in_dir, out_dir, symbol_list):
    symbol_passing_tiles = dict()
    for symbol in symbol_list:
        symbol_string = f'({symbol})'
        new_tile_type = copy.deepcopy(tile_type)
        for d in Dirs:
            if d == in_dir or d == out_dir:
                new_tile_type.sides[d].label = f'{symbol_string}{tile_type.sides[d].label}'
        new_tile_type.name = f'{symbol_string}{tile_type.name}'
        new_tile_type.label = f'{symbol_string}{tile_type.label}'
        #new_tile_type.Print()
        symbol_passing_tiles[new_tile_type.name] = new_tile_type

    return symbol_passing_tiles

def get_quine_builder_tile_set(num_bits_for_symbols, counter_width, b_output=False, b_use_p=False, b_include_fillers=True, b_make_square=True):
    quine_builder_tile_set = dict()

    phase_markers = ['ext', 'pad', 'null', 'ctr', 'blank', 'copy']
    curr_phase_count = 0

    # PHASE 1: Append templates for tiles of T_S
    #   Start condition: expect '!' marker on leftmost tile of seed row
    #   Loop over each tile in the seed row, moving '!' marker from left to right
    #       For each symbol found, append the template for a tile with that symbol to the right side of the row
    #   At right end ('B4'), erase '!', return to left, go to next phase

    curr_phase_marker = phase_markers[curr_phase_count]

    tile_type = PyTASTileType()
    tile_type.name = f'QuineStarter'
    tile_type.label = tile_type.name
    tile_type.color = 'green'

    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[S] = Glue(2, f'B0!')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_pass_B0!')

    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    tiles_to_find_bang = get_tiles_to_get_symbol_at_marker(curr_phase_marker, post_seed_row_symbol_list[1:], 'blue', W_to_E)
    quine_builder_tile_set.update(tiles_to_find_bang)
    draw_tile_set(tiles_to_find_bang, 'test', 'tiles_to_find_bang', b_output)

    curr_vert_symbols = copy.deepcopy(seed_row_symbol_list)
    curr_vert_symbols += ['ctr', '$$']
    curr_vert_symbols.remove('B0!')

    tiles_to_pass_bang_val = get_tiles_to_pass_symbols_right(curr_phase_marker, seed_row_symbol_list, curr_vert_symbols, 'B4', '$$', 'white')
    quine_builder_tile_set.update(tiles_to_pass_bang_val)
    draw_tile_set(tiles_to_pass_bang_val, 'test', 'tiles_to_pass_bang_val', b_output)

    #if False:
    #    seed_tile_template_tiles = get_seed_tile_template_tiles(curr_phase_marker, num_bits_for_symbols, 'green', b_compress=True)
    #else:
    seed_tile_template_tiles = get_seed_tile_template_tiles(curr_phase_marker, counter_width, 'green', b_compress=False)
    quine_builder_tile_set.update(seed_tile_template_tiles)
    draw_tile_set(seed_tile_template_tiles, 'test', 'seed_tile_template_tiles', b_output)

    left_reset_tiles = get_tiles_for_left_reset_with_prefix_check(curr_vert_symbols, 'B4', curr_phase_marker, '', 'white')

    # Get leftmost tile for this phase
    tile_type = PyTASTileType()
    tile_type.name = f'RtoL_{curr_phase_marker}_B0'
    tile_type.label = tile_type.name
    tile_type.color = 'white'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[E] = Glue(1, f'RtoL_ok_{curr_phase_marker}')
    IO_mark_tile_type(tile_type, [S,E])
    left_reset_tiles[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'LtoR_{curr_phase_marker}_B0'
    tile_type.label = tile_type.name
    tile_type.color = 'white'
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}')
    IO_mark_tile_type(tile_type, [S])
    left_reset_tiles[tile_type.name] = tile_type

    quine_builder_tile_set.update(left_reset_tiles)
    draw_tile_set(left_reset_tiles, 'test', 'left_reset_tiles', b_output)

    # Last extension of PHASE 1 when (west == 'ext') and (south == 'ext_$$') since that was the original location of 'B4'
     
    tile_type = PyTASTileType()
    tile_type.name = f'Cleanup_start_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(1, f'{curr_phase_marker}_$$')
    tile_type.sides[N] = Glue(1, f'$')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}')
    tile_type.sides[E] = Glue(1, f'Cleanup_{curr_phase_marker}')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    ext_cleanup_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols, [('$$', '$')], 'B4', curr_phase_marker, 'pink', W_to_E)

    # Modify the final tile to do the B4 extension
    # Final tile: south = B4, north = $, east = {curr_phase_marker}_B4 (strength-2)
    final_tile.sides[S] = Glue(1, f'B4')
    final_tile.sides[N] = Glue(1, f'$')
    final_tile.sides[W] = Glue(1, f'Cleanup_{curr_phase_marker}')
    final_tile.sides[E] = Glue(2, f'{curr_phase_marker}_B4')
    IO_mark_tile_type(final_tile, [S,W])
    quine_builder_tile_set.update(ext_cleanup_tiles)
    draw_tile_set(ext_cleanup_tiles, 'test', 'ext_cleanup_tiles', b_output)

    curr_vert_symbols.remove('$$')

    if b_use_p:
        ##############################################################################################################################################
        # PHASE 2: Replace 'p' symbols with padding width of 0's
        #   1. Mark first bit after 'B2' with '*'
        #   2. Find leftmost 'p':
        #       If found, insert '0' to its right, bumping everything to the right by one position until right side
        #       If no remaining 'p' symbols (i.e. reach 'B4' first), return to left and go to next phase
        #   3. Return to the left side ('B0')
        #   4. Go right to bit marked with '*' and remove '*':
        #       If tile to the right is '0', mark it
        #           If there is another '0' to the right of that, loop to step 2
        #           If that was the rightmost '0', loop to step 2 but remembering that the lefmost 'p' will be replaced with '0' (no insertion)

        # Initiate Phase 2 when arriving at 'B0' with 'RtoL_ext' instead of 'RtoL_ok_ext'
        tile_type = PyTASTileType()
        tile_type.name = f'Done_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[S] = Glue(1, f'B0')
        tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+1]}_B0')
        tile_type.sides[E] = Glue(1, f'RtoL_{curr_phase_marker}')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

        curr_phase_count += 1
        curr_phase_marker = phase_markers[curr_phase_count]

        tile_type = PyTASTileType()
        tile_type.name = f'Start_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0')
        tile_type.sides[N] = Glue(1, f'B0')
        tile_type.sides[E] = Glue(1, f'MS1AS2_{curr_phase_marker}_1')
        IO_mark_tile_type(tile_type, [S])
        quine_builder_tile_set[tile_type.name] = tile_type

        before_vert_symbols = ['B0', 'B1', 'B2', '0', '1', 'n', 'p', '#']
        mark_first_pad_bit_tiles = get_first_unmarked_sym1_after_sym2_tiles(curr_phase_marker, curr_phase_marker, '1', '2', before_vert_symbols, ['0'], 'B2', ['0'], 'purple')
        quine_builder_tile_set.update(mark_first_pad_bit_tiles)
        draw_tile_set(mark_first_pad_bit_tiles, 'test', 'mark_first_pad_bit_tiles', b_output)

        # Found first unmarked '0' after 'B2'. See if there's another '0' to its immediate right
        #   (A) If yes: Insert '0' to right of leftmost 'p', bumping everything to the right, then return left and repeat
        #   (B) If no: Replace leftmost 'p' with '0', return left resetting marked '0's to unmarked and repeat
        #   (C) If no and no remaining 'p's, return left and start next phase

        # Make the tiles that determine/handle if we're in case (A) or (B,C)
        #  Determine Case (A)
        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_A_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'0')
        tile_type.sides[N] = Glue(1, f'0')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

        # Handle Case (A)
        curr_vert_symbols_sans_B4 = copy.deepcopy(curr_vert_symbols)
        curr_vert_symbols_sans_B4.remove('B4')
        case_A_padding_tiles = get_tiles_to_find_symbol(f'{curr_phase_marker}_A', f'{curr_phase_marker}_A_ins_0', 'p', curr_vert_symbols_sans_B4, 'purple', W_to_E)
        quine_builder_tile_set.update(case_A_padding_tiles)
        draw_tile_set(case_A_padding_tiles, 'test', 'case_A_padding_tiles', b_output)

        case_A_insertion_tiles = get_right_shift_tiles(f'{curr_phase_marker}_A_ins', curr_vert_symbols, 'B4', 'orange')
        quine_builder_tile_set.update(case_A_insertion_tiles)
        draw_tile_set(case_A_insertion_tiles, 'test', 'case_A_insertion_tiles', b_output)

        # Make left reset tiles starting from pad_A_ins_B4
        pad_left_reset_tiles = get_tiles_for_left_reset(curr_vert_symbols + [f'{curr_phase_marker}_0'], 'B4', curr_phase_marker, 'pad_A_ins', 'white')
        quine_builder_tile_set.update(pad_left_reset_tiles)
        draw_tile_set(pad_left_reset_tiles, 'test', 'pad_left_reset_tiles', b_output)

        tile_type = PyTASTileType()
        tile_type.name = f'Continue_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'B0')
        tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
        tile_type.sides[E] = Glue(1, f'RtoL_{curr_phase_marker}')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

        #  Initiate Case (B,C)
        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_BC_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'B3')
        tile_type.sides[N] = Glue(1, f'B3')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_BC')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

        # Determine case B or C: if a 'p' is found, 'B', else 'C'
        case_BC_padding_tiles = get_tiles_to_find_symbol(f'{curr_phase_marker}_BC', f'{curr_phase_marker}_B', 'p', curr_vert_symbols, 'purple', W_to_E, '0')
        quine_builder_tile_set.update(case_BC_padding_tiles)
        draw_tile_set(case_BC_padding_tiles, 'test', 'case_BC_padding_tiles', b_output)

        case_B_right_tiles = get_tiles_to_go_right(f'{curr_phase_marker}_B', curr_vert_symbols, 'B4', f'{curr_phase_marker}_B', 'yellow')
        quine_builder_tile_set.update(case_B_right_tiles)
        draw_tile_set(case_B_right_tiles, 'test', 'case_BC_right_tiles', b_output)

        # Grow up from 'B4' to start cleanup
        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_BC_RIGHT'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B_B4')
        tile_type.sides[N] = Glue(1, f'B4')
        tile_type.sides[W] = Glue(1, f'Cleanup_{curr_phase_marker}_B')
        tile_type.sides[E] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [S])
        quine_builder_tile_set[tile_type.name] = tile_type

        swap_pairs = [(f'{curr_phase_marker}_0', '0')]
        case_B_left_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols + [f'{curr_phase_marker}_0'], swap_pairs, 'B0', f'{curr_phase_marker}_B', 'yellow', E_to_W)
        quine_builder_tile_set.update(case_B_left_tiles)
        draw_tile_set(case_B_left_tiles, 'test', 'case_B_left_tiles', b_output)

        # Grow up from 'B0' after cleanup
        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_BC_LEFT'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'B0')
        tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
        tile_type.sides[E] = Glue(1, f'Cleanup_{curr_phase_marker}_B')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

        # Case (C) occurs when no 'p' is found, so 'B4' is reached
        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_C_DONE'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[S] = Glue(1, f'B4')
        tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_C_B4')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A')
        tile_type.sides[E] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

        tile_type = PyTASTileType()
        tile_type.name = f'Pad_ABC1_C_CLEAN'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_C_B4')
        tile_type.sides[N] = Glue(1, 'B4')
        tile_type.sides[W] = Glue(1, f'Cleanup_{curr_phase_marker}_C')
        tile_type.sides[E] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [S])
        quine_builder_tile_set[tile_type.name] = tile_type

        pad_case_C_cleanup_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols + [f'{curr_phase_marker}_0'], [('$$', '$'), (f'{curr_phase_marker}_0', '0')], 'B0', f'{curr_phase_marker}_C', 'lime', E_to_W)
        quine_builder_tile_set.update(pad_case_C_cleanup_tiles)
        draw_tile_set(pad_case_C_cleanup_tiles, 'test', 'pad_case_C_cleanup_tiles', b_output)

        #curr_vert_symbols.remove('$$')
        curr_vert_symbols.remove('p')
        #print(f'curr_vert_symbols: {curr_vert_symbols}')
    else: # Not using 'p' symbol

        # Initiate Phase 2 when arriving at 'B0' with 'RtoL_ext' instead of 'RtoL_ok_ext'
        tile_type = PyTASTileType()
        tile_type.name = f'Done_{curr_phase_marker}'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[S] = Glue(1, f'B0')
        tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+2]}_B0')
        tile_type.sides[E] = Glue(1, f'RtoL_{curr_phase_marker}')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type


        #MJP TODO: need some cleanup tiles to swap '$$' to '$'
        # pad_case_C_cleanup_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols + [f'{curr_phase_marker}_0'], [('$$', '$'), (f'{curr_phase_marker}_0', '0')], 'B0', f'{curr_phase_marker}_C', 'lime', E_to_W)
        # quine_builder_tile_set.update(pad_case_C_cleanup_tiles)
        # draw_tile_set(pad_case_C_cleanup_tiles, 'test', 'pad_case_C_cleanup_tiles', b_output)

        # curr_vert_symbols.remove('$$')
        # curr_vert_symbols.remove('p')


        curr_phase_count += 2
        curr_phase_marker = phase_markers[curr_phase_count]


    ##############################################################################################################################################
    # PHASE 3: Replace 'n' symbols with ctr width of 0's
    #   1. Mark first bit (0 or 1) after 'B0' with 'null'
    #   2. Find leftmost 'n':
    #       If found, insert '0' to its right, bumping everything to the right by one position until right side
    #       If no remaining 'n' symbols (i.e. reach 'B4' first), return to left and go to next phase
    #   3. Return to the left side ('B0')
    #   4. Go right to bit (0 or 1) marked with 'null' and remove 'null':
    #       If tile to the right is '0' or '1', mark it
    #           If there is another '0' or '1' to the right of that, loop to step 2
    #           If that was the rightmost '0' or '1', loop to step 2 but remembering that the lefmost 'n' will be replaced with '0' and an
    #            insertion will happen (for the strength bit)

    if b_use_p:
        # Initiate Phase 3 when arriving at 'B0' with 'Cleanup_pad_C'
        tile_type = PyTASTileType()
        tile_type.name = f'Done_{curr_phase_marker}_C'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[S] = Glue(1, f'B0')
        tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+1]}_B0')
        tile_type.sides[E] = Glue(1, f'Cleanup_{curr_phase_marker}_C')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

        curr_phase_count += 1
        curr_phase_marker = phase_markers[curr_phase_count]

    tile_type = PyTASTileType()
    tile_type.name = f'Start_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'MFUS_{curr_phase_marker}_1')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    mark_first_null_bit_tiles = mark_first_unmarked(curr_phase_marker, curr_phase_marker, '1', '2', ['0', '1'], 'B1', 'pink')
    quine_builder_tile_set.update(mark_first_null_bit_tiles)
    draw_tile_set(mark_first_null_bit_tiles, 'test', 'mark_first_null_bit_tiles', b_output)

    # Found first unmarked '0' or '1' after 'B0'. See if there's another '0' or '1' to its immediate right
    #   (A) If yes: Insert '0' to right of leftmost 'n', bumping everything to the right, then return left and repeat
    #   (B) If no: Replace leftmost 'n' with '0' and insert '0', return left resetting marked '0's and '1's to unmarked and repeat
    #   (C) If no and no remaining 'n's, return left and start next phase

    # Make the tiles that determine/handle if we're in case (A) or (B,C)
    #  Determine Case (A)
    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_A0_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'0')
    tile_type.sides[N] = Glue(1, f'0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_2')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_A1_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'1')
    tile_type.sides[N] = Glue(1, f'1')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_2')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Handle Case (A)
    curr_vert_symbols_sans_B4 = copy.deepcopy(curr_vert_symbols)
    curr_vert_symbols_sans_B4.remove('B4')
    null_case_A_padding_tiles = get_tiles_to_find_symbol(f'{curr_phase_marker}_A', f'{curr_phase_marker}_A_ins_0', 'n', curr_vert_symbols_sans_B4, 'purple', W_to_E)
    quine_builder_tile_set.update(null_case_A_padding_tiles)
    draw_tile_set(null_case_A_padding_tiles, 'test', 'null_case_A_padding_tiles', b_output)

    null_case_A_insertion_tiles = get_right_shift_tiles(f'{curr_phase_marker}_A_ins', curr_vert_symbols, 'B4', 'orange')
    quine_builder_tile_set.update(null_case_A_insertion_tiles)
    draw_tile_set(null_case_A_insertion_tiles, 'test', 'null_case_A_insertion_tiles', b_output)

    # Make left reset tiles starting from null_A_ins_B4
    null_left_reset_tiles = get_tiles_for_left_reset(curr_vert_symbols + [f'{curr_phase_marker}_0', f'{curr_phase_marker}_1'], 'B4', curr_phase_marker, 'null_A_ins', 'white')
    quine_builder_tile_set.update(null_left_reset_tiles)
    draw_tile_set(null_left_reset_tiles, 'test', 'null_left_reset_tiles', b_output)

    tile_type = PyTASTileType()
    tile_type.name = f'Continue_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[E] = Glue(1, f'RtoL_{curr_phase_marker}')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    #  Initiate Case (B,C)
    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_BC_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'B1')
    tile_type.sides[N] = Glue(1, f'B1')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_BC')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_2')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Determine case B or C: if an 'n' is found, 'B', else 'C'
    null_case_BC_tiles = get_tiles_to_find_symbol(f'{curr_phase_marker}_BC', f'{curr_phase_marker}_B_ins_0', 'n', curr_vert_symbols_sans_B4, 'purple', W_to_E, '0')
    quine_builder_tile_set.update(null_case_BC_tiles)
    draw_tile_set(null_case_BC_tiles, 'test', 'null_case_BC_tiles', b_output)

    null_case_B_insertion_tiles = get_right_shift_tiles(f'{curr_phase_marker}_B_ins', curr_vert_symbols, 'B4', 'orange')
    quine_builder_tile_set.update(null_case_B_insertion_tiles)
    draw_tile_set(null_case_B_insertion_tiles, 'test', 'null_case_B_insertion_tiles', b_output)

    # Grow up from 'B4' to start cleanup
    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_BC_RIGHT'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B_ins_B4')
    tile_type.sides[N] = Glue(1, f'B4')
    tile_type.sides[W] = Glue(1, f'Cleanup_{curr_phase_marker}_B')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    swap_pairs = [(f'{curr_phase_marker}_0', '0'), (f'{curr_phase_marker}_1', '1')]
    null_case_B_left_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols + [f'{curr_phase_marker}_0', f'{curr_phase_marker}_1'], swap_pairs, 'B0', f'{curr_phase_marker}_B', 'yellow', E_to_W)
    quine_builder_tile_set.update(null_case_B_left_tiles)
    draw_tile_set(null_case_B_left_tiles, 'test', 'null_case_B_left_tiles', b_output)

    # Grow up from 'B0' after cleanup
    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_BC_LEFT'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[E] = Glue(1, f'Cleanup_{curr_phase_marker}_B')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Case (C) occurs when no 'n' is found, so 'B4' is reached
    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_C_DONE'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(1, f'B4')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_C_B4')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Null_ABC1_C_CLEAN'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_C_B4')
    tile_type.sides[N] = Glue(1, 'B4')
    tile_type.sides[W] = Glue(1, f'Cleanup_{curr_phase_marker}_C')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    null_case_C_cleanup_tiles, final_tile = get_cleanup_tiles(curr_vert_symbols + [f'{curr_phase_marker}_0', f'{curr_phase_marker}_1'], [(f'{curr_phase_marker}_0', '0'), (f'{curr_phase_marker}_1', '1')], 'B0', f'{curr_phase_marker}_C', 'lime', E_to_W)
    quine_builder_tile_set.update(null_case_C_cleanup_tiles)
    draw_tile_set(null_case_C_cleanup_tiles, 'test', 'null_case_C_cleanup_tiles', b_output)

    ##############################################################################################################################################
    # PHASE 4: Replace leftmost two 'ctr' symbols with current counter value, then increment it and repeat

    # Initiate Phase 4 when arriving at 'B0' with 'Cleanup_null_C'
    tile_type = PyTASTileType()
    tile_type.name = f'Done_{curr_phase_marker}_C'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+1]}_B0')
    tile_type.sides[E] = Glue(1, f'Cleanup_{curr_phase_marker}_C')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    curr_phase_count += 1
    curr_phase_marker = phase_markers[curr_phase_count]

    tile_type = PyTASTileType()
    tile_type.name = f'Start_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'MFUS_{curr_phase_marker}_1')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    mark_first_ctr_bit_tiles = mark_first_unmarked(curr_phase_marker, curr_phase_marker, '1', 'A', ['0', '1'], 'B1', 'pink', b_append_value=True)
    quine_builder_tile_set.update(mark_first_ctr_bit_tiles)
    draw_tile_set(mark_first_ctr_bit_tiles, 'test', 'mark_first_ctr_bit_tiles', b_output)

    # Look for unmarked bit between 'B0' and 'B1':
    #    (A) If one exists, 
    #        (i) If a 'ctr' exists to the right, then there will be two and insert that bit to right of two leftmost 'ctr's,
    #            bumping everything to the right (twice), then return left and repeat
    #        (ii) If no 'ctr' exists to the right, reset left, clearing mark on ctr bit and start next phase
    #    (B) If none exists, replace the leftmost two 'ctr's with 1s, then return left and increment

    # Was able to mark a bit, so go right to find a 'ctr'
    for bit in ['0', '1']:
        for vert_symbol in ['B1', 'B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr']:
            tile_type = PyTASTileType()
            tile_type.name = f'Ctr_A_{curr_phase_marker}_{bit}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = 'purple'
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            if vert_symbol == 'ctr':
                tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_ins_{bit}_{bit}')
            else:
                tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_{bit}')
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_{bit}')
            IO_mark_tile_type(tile_type, [S,W])
            quine_builder_tile_set[tile_type.name] = tile_type

    # No bit to mark
    tile_type = PyTASTileType()
    tile_type.name = f'AllMarked_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'red'
    tile_type.sides[S] = Glue(1, f'B1')
    tile_type.sides[N] = Glue(1, f'B1')
    tile_type.sides[W] = Glue(1, f'MFUS_{curr_phase_marker}_1')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_1')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Find leftmost ctr and change it to 1
    for vert_symbol in ['B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_B_1_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        if vert_symbol == 'ctr':
            tile_type.sides[N] = Glue(1, f'1') # Set glue strength bit to 1
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_2')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_1')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_1')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

    # Find second leftmost ctr and change it to 1
    for vert_symbol in ['B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_B_2_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        if vert_symbol == 'ctr':
            tile_type.sides[N] = Glue(1, f'1') # Set glue strength bit to 1
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_3')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_2')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_2')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

    # Done with this pair of counters, go to the right, then back left, to increment counter
    for vert_symbol in ['B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_B_3_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_3')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_3')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

    # At right side
    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_3_{curr_phase_marker}_B4'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'B4')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B_3')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_3')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_4_{curr_phase_marker}_B4'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(2,f'{curr_phase_marker}_B_3')
    tile_type.sides[N] = Glue(1, 'B4')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_4')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S,])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Go back to the left
    for vert_symbol in ['B1', 'B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr', 'ctr_0', 'ctr_1']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_B_4_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        if vert_symbol in ['ctr_0', 'ctr_1']:
            tile_type.sides[N] = Glue(1, f'{vert_symbol[-1]}')
        else:
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_4')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_4')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

    # At the left end, initiate counter increment
    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_4_{curr_phase_marker}_B0'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B_5')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_4')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_B0'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B_5')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_a')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Tiles to increment the counter
        # Change 1s to 0s until finding first 0 - change it to 1 then done
    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_0_a'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[N] = Glue(1, f'1')
    tile_type.sides[S] = Glue(1, f'0')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_a')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_b')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_0_b'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[N] = Glue(1, f'0')
    tile_type.sides[S] = Glue(1, f'0')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_b')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_b')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_1_a'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[N] = Glue(1, f'0')
    tile_type.sides[S] = Glue(1, f'1')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_a')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_a')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_1_b'
    tile_type.label = tile_type.name
    tile_type.color = 'purple'
    tile_type.sides[N] = Glue(1, f'1')
    tile_type.sides[S] = Glue(1, f'1')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_b')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_b')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type
    
    # Tiles to go to the right
    for vert_symbol in ['B1', 'B2', 'B3', '$', '|1', '|2', '#', '0', '1', 'ctr']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_B_5_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'purple'
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B_5_b')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_b')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

    # At right end, hook up with previous tiles that come back left and restart the process
    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_B_5_{curr_phase_marker}_B4'
    tile_type.label = tile_type.name
    tile_type.color = 'blue'
    tile_type.sides[S] = Glue(1, f'B4')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B4')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_B_5_b')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S,W])
    quine_builder_tile_set[tile_type.name] = tile_type

    # # Make the tiles that determine/handle if we're in case (A) or (B,C)
    # #  Determine Case (A)
    # for bit in ['0', '1']:
    #     tile_type = PyTASTileType()
    #     tile_type.name = f'Ctr_ABC1_A0_{curr_phase_marker}_{bit}'
    #     tile_type.label = tile_type.name
    #     tile_type.color = 'purple'
    #     tile_type.sides[S] = Glue(1, f'0')
    #     tile_type.sides[N] = Glue(1, f'0')
    #     tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_{bit}')
    #     tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_2_{bit}')
    #     IO_mark_tile_type(tile_type, [S,W])
    #     quine_builder_tile_set[tile_type.name] = tile_type

    #     tile_type = PyTASTileType()
    #     tile_type.name = f'Ctr_ABC1_A1_{curr_phase_marker}_{bit}'
    #     tile_type.label = tile_type.name
    #     tile_type.color = 'purple'
    #     tile_type.sides[S] = Glue(1, f'1')
    #     tile_type.sides[N] = Glue(1, f'1')
    #     tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_{bit}')
    #     tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_2_{bit}')
    #     IO_mark_tile_type(tile_type, [S,W])
    #     quine_builder_tile_set[tile_type.name] = tile_type

    ctr_vert_symbols = ['$', '|1', '|2', '#', '0', '1', 'ctr']
    curr_vert_symbols_sans_ctr = copy.deepcopy(ctr_vert_symbols)
    curr_vert_symbols_sans_ctr.remove('ctr')

    # Handle Case (A)
    for bit in ['0', '1']:
        # ctr_case_A_tiles = get_tiles_to_find_symbol(f'{curr_phase_marker}_A_{bit}', f'{curr_phase_marker}_A_ins_{bit}_{bit}', 'ctr', curr_vert_symbols_sans_B4, 'purple', W_to_E)
        # quine_builder_tile_set.update(ctr_case_A_tiles)
        # draw_tile_set(ctr_case_A_tiles, 'test', 'ctr_case_A_tiles', b_output)

        ctr_case_A_insertion_tiles = get_right_shift_tiles(f'{curr_phase_marker}_A_ins_{bit}', curr_vert_symbols_sans_ctr, 'B4', 'orange')
        quine_builder_tile_set.update(ctr_case_A_insertion_tiles)
        draw_tile_set(ctr_case_A_insertion_tiles, 'test', 'ctr_case_A_insertion_tiles', b_output)

        # Upon hitting the second 'ctr' and carrying an aribtrary vert_symbol, make the tiles that cause 'ctr' '0' to be dropped off, then two values carried forward
        for vert_symbol in curr_vert_symbols_sans_ctr:
            tile_type = PyTASTileType()
            tile_type.name = f'Ctr_2nd_{curr_phase_marker}_{bit}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = 'blue'
            tile_type.sides[S] = Glue(1, f'ctr')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_ins2_{bit}')
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_ins_{bit}_{vert_symbol}')
            IO_mark_tile_type(tile_type, [S,W])
            quine_builder_tile_set[tile_type.name] = tile_type

            tile_type = PyTASTileType()
            tile_type.name = f'Ctr_2nd2_{curr_phase_marker}_{bit}_{vert_symbol}'
            tile_type.label = tile_type.name
            tile_type.color = 'blue'
            tile_type.sides[S] = Glue(1, f'{vert_symbol}')
            tile_type.sides[N] = Glue(1, f'ctr')
            tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_ins3_{bit}_{vert_symbol}')
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_ins2_{bit}')
            IO_mark_tile_type(tile_type, [S,W])
            quine_builder_tile_set[tile_type.name] = tile_type
            
            for vert_symbol2 in ctr_vert_symbols:
                tile_type = PyTASTileType()
                tile_type.name = f'Ctr_2nd3_{curr_phase_marker}_{bit}_{vert_symbol}_{vert_symbol2}'
                tile_type.label = tile_type.name
                tile_type.color = 'blue'
                tile_type.sides[S] = Glue(1, f'{vert_symbol2}')
                tile_type.sides[N] = Glue(1, f'{bit}')
                tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_ins4_{vert_symbol}_{vert_symbol2}')
                tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_ins3_{bit}_{vert_symbol}')
                IO_mark_tile_type(tile_type, [S,W])
                quine_builder_tile_set[tile_type.name] = tile_type

    for vert_symbol in ctr_vert_symbols:
        for vert_symbol2 in ctr_vert_symbols:
            for vert_symbol3 in ctr_vert_symbols:
                tile_type = PyTASTileType()
                tile_type.name = f'Ctr_2nd4_{curr_phase_marker}_{vert_symbol}_{vert_symbol2}_{vert_symbol3}'
                tile_type.label = tile_type.name
                tile_type.color = 'blue'
                tile_type.sides[S] = Glue(1, f'{vert_symbol3}')
                tile_type.sides[N] = Glue(1, f'{vert_symbol}')
                tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_A_ins4_{vert_symbol2}_{vert_symbol3}')
                tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_ins4_{vert_symbol}_{vert_symbol2}')
                IO_mark_tile_type(tile_type, [S,W])
                quine_builder_tile_set[tile_type.name] = tile_type

            tile_type = PyTASTileType()
            tile_type.name = f'Ctr_2nd_right1_{curr_phase_marker}_{vert_symbol}_{vert_symbol2}'
            tile_type.label = tile_type.name
            tile_type.color = 'blue'
            tile_type.sides[S] = Glue(1, f'B4')
            tile_type.sides[N] = Glue(1, f'{vert_symbol}')
            tile_type.sides[E] = Glue(2, f'{curr_phase_marker}_A_ins4_r1_{vert_symbol2}')
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_ins4_{vert_symbol}_{vert_symbol2}')
            IO_mark_tile_type(tile_type, [S,W])
            quine_builder_tile_set[tile_type.name] = tile_type

        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_2nd_right2_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = 'blue'
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[E] = Glue(2, f'{curr_phase_marker}_A_ins4_r2')
        tile_type.sides[W] = Glue(2, f'{curr_phase_marker}_A_ins4_r1_{vert_symbol}')
        tile_type.sides[S] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [W])
        quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_2nd_right3_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = 'blue'
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B4')
    tile_type.sides[W] = Glue(2, f'{curr_phase_marker}_A_ins4_r2')
    tile_type.sides[E] = Glue(1, f'FILL')
    tile_type.sides[S] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [W])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Go back to the left to get the next ctr bit
    aqua_color = 'rgb(0,255,255)'

    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_left_reset_B4_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = aqua_color
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B4')
    tile_type.sides[N] = Glue(1, f'B4')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_reset')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    for vert_symbol in ['B1', 'B2', 'B3', 'B4', '$', '|1', '|2', '#', '0', '1', 'ctr', 'ctr_0', 'ctr_1']:
        tile_type = PyTASTileType()
        tile_type.name = f'Ctr_left_reset_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = aqua_color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_reset')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_reset')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

    # Hit left side, start a new row
    tile_type = PyTASTileType()
    tile_type.name = f'Ctr_left_end_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = aqua_color
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_reset')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type


    ##############################################################################################################################################
    # PHASE 5: Change values B0-B1, and values B2-B3, to '_'

    for bit in ['0', '1']:
        tile_type = PyTASTileType()
        tile_type.name = f'End_{curr_phase_marker}_{bit}'
        tile_type.label = tile_type.name
        tile_type.color = 'red'
        tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+1]}_B4')
        tile_type.sides[S] = Glue(1, f'B4')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_A_{bit}')
        tile_type.sides[E] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [S,W])
        quine_builder_tile_set[tile_type.name] = tile_type

    curr_phase_count += 1
    curr_phase_marker = phase_markers[curr_phase_count]

    pink_color = 'rgb(255,0,255)'

    tile_type = PyTASTileType()
    tile_type.name = f'RtoL_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = pink_color
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B4')
    tile_type.sides[N] = Glue(1, f'B4')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean')
    tile_type.sides[E] = Glue(1, f'FILL')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Continue left until B3
    for vert_symbol in ['B3', 'B4', '$', '|1', '|2', '#', '0', '1']:
        tile_type = PyTASTileType()
        tile_type.name = f'Blank_RtoL_{curr_phase_marker}_{vert_symbol}'
        tile_type.label = tile_type.name
        tile_type.color = aqua_color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        if vert_symbol == 'B3':
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_1')
        else:
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

    # Blank out between B3 and B2 (which will only consist of 0s)
    tile_type = PyTASTileType()
    tile_type.name = f'RtoL_{curr_phase_marker}_B3_B2'
    tile_type.label = tile_type.name
    tile_type.color = pink_color
    tile_type.sides[S] = Glue(1, f'0')
    tile_type.sides[N] = Glue(1, f'_')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_1')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean_1')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Transition at B2
    tile_type = PyTASTileType()
    tile_type.name = f'RtoL_{curr_phase_marker}_B2'
    tile_type.label = tile_type.name
    tile_type.color = pink_color
    tile_type.sides[S] = Glue(1, f'B2')
    tile_type.sides[N] = Glue(1, f'B2')
    tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_2')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean_1')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    # Continue left until B1
    for vert_symbol in ['B1', '$', '|1', '|2', '#', '0', '1']:
        tile_type = PyTASTileType()
        tile_type.name = f'Blank_RtoL_{curr_phase_marker}_{vert_symbol}_2'
        tile_type.label = tile_type.name
        tile_type.color = aqua_color
        tile_type.sides[S] = Glue(1, f'{vert_symbol}')
        tile_type.sides[N] = Glue(1, f'{vert_symbol}')
        if vert_symbol == 'B1':
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_3')
        else:
            tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_2')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean_2')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type

    # Blank out between B1 and B0 (which will only consist of (possibly marked) 0s and 1s)
    for bit in ['0', '1', 'ctr_0', 'ctr_1']:
        tile_type = PyTASTileType()
        tile_type.name = f'RtoL_{curr_phase_marker}_{bit}_B1_B0'
        tile_type.label = tile_type.name
        tile_type.color = pink_color
        tile_type.sides[S] = Glue(1, f'{bit}')
        tile_type.sides[N] = Glue(1, f'_')
        tile_type.sides[W] = Glue(1, f'{curr_phase_marker}_left_clean_3')
        tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean_3')
        IO_mark_tile_type(tile_type, [S,E])
        quine_builder_tile_set[tile_type.name] = tile_type


    ##############################################################################################################################################
    # PHASE 6: Copy B0-B3 to B4-B7

    copy_colors = ['rgb(0,255,0)', 'rgb(0,123,255)', 'rgb(255,123,0)']

    tile_type = PyTASTileType()
    tile_type.name = f'Clean_left_end_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[0]
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{phase_markers[curr_phase_count+1]}_B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_left_clean_3')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    curr_phase_count += 1
    curr_phase_marker = phase_markers[curr_phase_count]

    tile_type = PyTASTileType()
    tile_type.name = f'Copy_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[0]
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B2B3_RGNCPY1')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    vert_symbol_list = ['B0', 'B1', 'B2', 'B3', 'B4', '$', '|1', '|2', '#', '0', '1', '_']
    copy_B2B3_tiles = get_tiles_to_copy_region_to_right(f'{curr_phase_marker}_B2B3', 'B2', 'B3', 'B4', 'B5', ['_'], vert_symbol_list, copy_colors[0])
    quine_builder_tile_set.update(copy_B2B3_tiles)
    draw_tile_set(copy_B2B3_tiles, 'test', 'copy_B2B3_tiles', b_output)

    tile_type = PyTASTileType()
    tile_type.name = f'Copy_B2B3_Done_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[1]
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B1B2_B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B2B3_RGNCPY_DONE')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Copy_B2B3_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[1]
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B1B2_B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B1B2_RGNCPY1')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    vert_symbol_list = ['B0', 'B1', 'B2', 'B3', 'B4', 'B5', '$', '|1', '|2', '#', '0', '1', '_']
    copy_B1B2_tiles = get_tiles_to_copy_region_to_right(f'{curr_phase_marker}_B1B2', 'B1', 'B2', 'B5', 'B6', ['#', '0', '1'], vert_symbol_list, copy_colors[1])
    quine_builder_tile_set.update(copy_B1B2_tiles)
    draw_tile_set(copy_B1B2_tiles, 'test', 'copy_B1B2_tiles', b_output)

    tile_type = PyTASTileType()
    tile_type.name = f'Copy_B1B2_Done_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[1]
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(2, f'{curr_phase_marker}_B0B1_B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B1B2_RGNCPY_DONE')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    tile_type = PyTASTileType()
    tile_type.name = f'Copy_B0B1_{curr_phase_marker}'
    tile_type.label = tile_type.name
    tile_type.color = copy_colors[1]
    tile_type.sides[S] = Glue(2, f'{curr_phase_marker}_B0B1_B0')
    tile_type.sides[N] = Glue(1, f'B0_SRC')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B0B1_RGNCPY1')
    IO_mark_tile_type(tile_type, [S])
    quine_builder_tile_set[tile_type.name] = tile_type

    vert_symbol_list = ['B0', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6', '$', '|1', '|2', '#', '0', '1', '_']
    copy_B0B1_tiles = get_tiles_to_copy_region_to_right(f'{curr_phase_marker}_B0B1', 'B0', 'B1', 'B6', 'B7', ['_'], vert_symbol_list, copy_colors[2])
    quine_builder_tile_set.update(copy_B0B1_tiles)
    draw_tile_set(copy_B0B1_tiles, 'test', 'copy_B0B1_tiles', b_output)

    tile_type = PyTASTileType()
    tile_type.name = f'Quine_Complete'
    tile_type.label = tile_type.name
    tile_type.color = 'green'
    tile_type.sides[S] = Glue(1, f'B0')
    tile_type.sides[N] = Glue(1, f'B0')
    tile_type.sides[E] = Glue(1, f'{curr_phase_marker}_B0B1_RGNCPY_DONE')
    # Glue to hook up with squaring tiles
    tile_type.sides[W] = Glue(2, f'west*')
    IO_mark_tile_type(tile_type, [S,E])
    quine_builder_tile_set[tile_type.name] = tile_type

    ##############################################################################################################################################
    # MACROTILE BUILDING
    # PHASE 7: add filler tiles and make into a square with proper exterior I/O markings

    if b_include_fillers:
        fill_color = 'rgb(123,123,123)'

        tile_type = PyTASTileType()
        tile_type.name = f'Ext_FILL'
        tile_type.label = tile_type.name
        tile_type.color = fill_color
        tile_type.sides[N] = Glue(2, f'Ext_FILL')
        tile_type.sides[E] = Glue(1, f'FILL_BOTT')
        IO_mark_tile_type(tile_type, [N])
        quine_builder_tile_set[tile_type.name] = tile_type

        tile_type = PyTASTileType()
        tile_type.name = f'FILL'
        tile_type.label = tile_type.name
        tile_type.color = fill_color
        tile_type.sides[N] = Glue(1, f'FILL')
        tile_type.sides[E] = Glue(1, f'FILL')
        tile_type.sides[S] = Glue(1, f'FILL')
        tile_type.sides[W] = Glue(1, f'FILL')
        IO_mark_tile_type(tile_type, [N,W])
        quine_builder_tile_set[tile_type.name] = tile_type

        tile_type = PyTASTileType()
        tile_type.name = f'FILL_BOTTOM'
        tile_type.label = tile_type.name
        tile_type.color = fill_color
        tile_type.sides[N] = Glue(1, f'FILL')
        tile_type.sides[E] = Glue(1, f'FILL_BOTT')
        #tile_type.sides[S] = Glue(1, f'FILL')
        tile_type.sides[W] = Glue(1, f'FILL_BOTT')
        IO_mark_tile_type(tile_type, [N,W])
        quine_builder_tile_set[tile_type.name] = tile_type
 
    if b_make_square:
        #  Get the tile type to make the assembly into a square (with proper exterior I/O markings)
        squaring_ttypes = gen_square_tile_set()
        print(f'Generated {len(squaring_ttypes)} squaring tile types')

        updated_squaring_types_dict = fix_IO_markings(squaring_ttypes)
        quine_builder_tile_set.update(updated_squaring_types_dict)


        # HACK: Tile to hook up with squaring tiles (for now)
        tile_type = PyTASTileType()
        tile_type.name = f'Squaring_init'
        tile_type.label = tile_type.name
        tile_type.color = 'rgb(255,50,255)'
        tile_type.sides[N] = Glue(2, f'FA*')
        tile_type.sides[E] = Glue(2, f'west*')
        tile_type.sides[S] = Glue(1, f'FA')
        IO_mark_tile_type(tile_type, [E])
        quine_builder_tile_set[tile_type.name] = tile_type


    return quine_builder_tile_set

def create_macrotile_test(macrotile_height, base_file_name, output_directory):
    
    tile_system = create_dummy_seed(macrotile_height)
    tile_system.tileTypes = fix_IO_markings(tile_system.tileTypes.values())

    squaring_ttypes = gen_square_tile_set()
    print(f'Generated {len(squaring_ttypes)} squaring tile types')

    updated_squaring_types_dict = fix_IO_markings(squaring_ttypes)
    tile_system.tileTypes.update(updated_squaring_types_dict)      

    # If output directory does not exist, create it!!
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    base_file_name = f'{output_directory}/{base_file_name}'
    
    tile_system.SaveToFile(base_file_name=base_file_name)
    print(f'Final tile system saved with {len(tile_system.tileTypes)} tile types and a seed of size {len(tile_system.seedPatch)}.')
    exit()


def color_tile_types(tile_set_dict, color):
    for tile_type_name in tile_set_dict:
        tile_set_dict[tile_type_name].color = color

def IO_mark_tile_type(tile_type, input_sides, special_output_sides=[]):
    for d in Dirs:
        if tile_type.sides[d].strength > 0:
            if d in special_output_sides:
                tile_type.sides[d].label += SPECIAL_OUTPUT_MARKERS[d]
            else:
                if d in input_sides:
                    tile_type.sides[d].label += INPUT_MARKERS[d]
                else:
                    tile_type.sides[d].label += OUTPUT_MARKERS[d]

def get_glue_analysis(tile_set_dict):
    glue_sets = get_glue_sets(tile_set_dict)
    num_glues_in_dir = {}
    #glue_log_lengths = {}
    for d in Dirs:
        num_glues_in_dir[d] = len(glue_sets[d])
        #glue_log_lengths[d] = math.ceil(math.log(num_glues_in_dir[d], 2))
    print(f'num_glues_in_dir: {num_glues_in_dir}')#, log lengths: {glue_log_lengths}')
    return num_glues_in_dir

# def get_input_sides(tile_type):
#     input_sides = []
#     for d in Dirs:
#         if (tile_type.sides[d].strength > 0) and (tile_type.sides[d].label.endswith(INPUT_MARKERS[d])):
#             input_sides.append(d)
#     return input_sides

# def get_output_sides(tile_type, b_include_special=True):
#     output_sides = []
#     for d in Dirs:
#         if tile_type.sides[d].strength > 0:
#             if tile_type.sides[d].label.endswith(OUTPUT_MARKERS[d]):
#                 output_sides.append(d)
#             elif tile_type.sides[d].label.endswith(SPECIAL_OUTPUT_MARKERS[d]):
#                 output_sides.append(d)
#     return output_sides

def get_umatched_glues(tile_set_dict):
    glue_sets = get_glue_sets_for_dirs(tile_set_dict)

    unmatched_N = glue_sets[N] - glue_sets[S]
    unmatched_E = glue_sets[E] - glue_sets[W]
    unmatched_S = glue_sets[S] - glue_sets[N]
    unmatched_W = glue_sets[W] - glue_sets[E]
    unmatched_glues = [unmatched_N, unmatched_E, unmatched_S, unmatched_W]

    return unmatched_glues

def get_seed_row_analysis(seed_row_encoding_list, glue_rep_width):
    number_seed_row_tiles = len(seed_row_encoding_list)
    padded_counter_value = seed_row_encoding_list[seed_row_encoding_list.index('B0!')+1 : seed_row_encoding_list.index('B1')]
    padded_counter_value_len = len(padded_counter_value)
    padding_value = seed_row_encoding_list[seed_row_encoding_list.index('B2')+1 : seed_row_encoding_list.index('B3')]
    padding_value_len = len(padding_value)
    tile_template_len = 20 + glue_rep_width

    p_count = 0
    n_count = 0
    for sym in seed_row_encoding_list:
        if sym == 'p':
            p_count += 1
        elif sym == 'n':
            n_count += 1

    print(f'number_seed_row_tiles: {number_seed_row_tiles}, padded_counter_value_len: {padded_counter_value_len}, padding_value_len: {padding_value_len}, n_count: {n_count}, p_count: {p_count}')

    phase_1_width_addition = number_seed_row_tiles * tile_template_len
    phase_1_height_addition = number_seed_row_tiles * 2
    print(f'Phase 1 adds witdh {phase_1_width_addition} and height {phase_1_height_addition}')



if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "hi:o:d:stfm:ru",
                            ["help", "in=", "out=", "dir=", "seed", "test", "fractal", "macro=", "rect", "iu="])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()

    input_tds_file_name = None
    base_file_name = None
    output_directory = '.'
    b_make_seed_row_tiles = False
    b_use_fake_seed_row = False
    macrotile_height = None
    b_compress_with_p = True
    b_make_square = True
    b_make_iu = False
    zeta_dim = 2
    epsilon = 0


    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-i', '--in']:
            input_tds_file_name = p
        elif o in ['-u', '--iu']:
            b_make_iu = True
            dim_info = p.split(',')
            zeta_dim = float(dim_info[0])
            epsilon = float(dim_info[1])
        elif o in ['-o', '--out']:
            base_file_name = p
        elif o in ['-d', '--dir']:
            output_directory = p
        elif o in ['-s', '--seed']:
            b_make_seed_row_tiles = True
        elif o in ['-t', '--test']:
            b_use_fake_seed_row = True
        elif o in ['-m', '--macro']:
            macrotile_height = int(p)
        elif o in ['-r', '--rect']:
            b_make_square = False
        else:
            DisplayHelp()

    if b_make_iu and (input_tds_file_name is not None):
        print(f"\nERROR: Only one option from 'in' or 'iu' should be used.")
        DisplayHelp()


    #########################################################################################################################################
    # Overview:
    # - read IU tile set, T_U
    # - guess values for the width of the glue counter (c_w) and the compressed width of glue representations (g_w)
    # - these will yield the width of padding p_w = c_w - g_w
    # - using g_w, generate the quine tile set T_q
    #   - g_w will slightly impact the size of T_q by determining how many tiles must be hard-coded for seed row tile symbols
    # - get the count of unique horizontal glue labels, g_h, and unique vertical glue labels, g_v, in T_q
    # - if g_w is not wide enough to accommodate either of g_h or g_v, increase g_w by 1 and redo
    # - eventually g_w can accommodate both g_h and g_v (this must happen since any increase cannot cause a doubling of glues)
    # - using g_w, the guessed value for c_w, and p_w, for T' = T_q \cup T_U, generate the tile set for the seed row, T_S, which encodes T'
    #   - let g_h' = |T_S| - 1 (i.e. the number of horizontal glues holding the seed row together)
    #   - if log(g_h' + g_h) > c_w, increase c_w by 1 and re-generate T_S
    #   - eventually, c_w will be large enough (this must happen since any increase only causes an additional tile for each of
    #     - the counter width c_w,  and
    #     - the padding width p_w)
    # - let the final tile set T = T' \cup T_S
    #########################################################################################################################################

    # if input_tds_file_name is None:
    #     print(f'ERROR: an input tds file must be specified.')
    #     quit()
    # else:

    if not macrotile_height is None:
        create_macrotile_test(macrotile_height, base_file_name, output_directory)
        exit()

    else:

        tile_system = PyTASTileSystem()
        iu_tile_set = {}
        if input_tds_file_name is not None:
            full_input_tds_file_name = f'{output_directory}/{input_tds_file_name}'
            tilesetLoadRet = tile_system.LoadTileSet(full_input_tds_file_name)
            print(f'Loaded {len(tile_system.tileTypes)} IU tile types from {full_input_tds_file_name}')
            iu_tile_set = copy.deepcopy(tile_system.tileTypes)
            tile_system.tileTypes = dict()

        elif b_make_iu:
            print(f'WARNING: Currently, the setting of zeta-dimension and epsilong are not supported.')
            #print(f'z: {zeta_dim}, epsilon: {epsilon}')
            iu_tile_list = gen_iu_ttypes()
            print(f'Generated {len(iu_tile_list)} IU tile types.')

            iu_tile_set = dict()
            for ttype in iu_tile_list:
                if ttype.name in iu_tile_set:
                    print(f'duplicate tile type name: {ttype.name}')
                    exit()
                iu_tile_set[ttype.name] = ttype

        if len(iu_tile_set) > 0:
            new_iu_tile_set = dict()
            for ttype_name in iu_tile_set:
                ttype = iu_tile_set[ttype_name]
                iu_name = f'IU_{ttype.name}'
                if iu_name in iu_tile_set:
                    print(f'duplicate tile type name: {iu_name}')
                    exit()
                ttype.name = iu_name
                new_iu_tile_set[iu_name] = ttype
            iu_tile_set = copy.deepcopy(new_iu_tile_set)
            tile_system.tileTypes = copy.deepcopy(iu_tile_set)          

        
        # unmatched_glue_sets = get_umatched_glues(tile_set_dict)
        # for unmatched_glue_set in unmatched_glue_sets:
        #     if len(unmatched_glue_set) > 0:
        #         print(f'WARNING: Unmatched glues:')
        #         print(f'unmatched_glue_sets: {unmatched_glue_sets}')


        num_seed_row_symbols = len(seed_row_symbol_list)
        num_bits_for_seed_row_symbols = len(bin(num_seed_row_symbols)[2:])
        print(f'There are {num_seed_row_symbols} seed row symbols, which require {num_bits_for_seed_row_symbols} bits to repreent.')

        glue_rep_width = num_bits_for_seed_row_symbols
        counter_width = glue_rep_width + 1

        b_final = False
        b_continue = True
        while b_continue:
            quine_builder_tile_set = get_quine_builder_tile_set(glue_rep_width, counter_width, b_output=False, b_use_p=b_compress_with_p, b_include_fillers=True, b_make_square=b_make_square) #Could tie this to b_final if drawing output is desired for final versions
            combined_tile_set = copy.deepcopy(iu_tile_set)
            combined_tile_set.update(quine_builder_tile_set)

            print(f'DEBUG: len(quine_builder_tile_set): {len(quine_builder_tile_set)}, len(iu_tile_set): {len(iu_tile_set)}, len(combined_tile_set): {len(combined_tile_set)}')

            # if b_use_fake_seed_row:
            #     temp_T = iu_tile_set
            # else:
            #     temp_T = T

            # Determine if glue_rep_width was large enough by checking the numbers of unique horizontal and vertical glues
            horizontal_glue_set = get_horizontal_glue_set(combined_tile_set)
            vertical_glue_set = get_vertical_glue_set(combined_tile_set)
            max_glue_set_len = max(len(horizontal_glue_set), len(vertical_glue_set))
            max_glue_value = max_glue_set_len - 1
            max_glue_value_len = len(bin(max_glue_value)[2:])
            print(f'trying glue_rep_width {glue_rep_width}, len horizontal_glue_set: {len(horizontal_glue_set)}, len vertical_glue_set: {len(vertical_glue_set)}, max_glue_value: {max_glue_value}, max_glue_value_len: {max_glue_value_len}')

            if max_glue_value_len > glue_rep_width:
                glue_rep_width += 1
            else:
                b_continue = False
                print(f'Stopping loop with glue_rep_width: {glue_rep_width}')
                
                if not b_final:
                    counter_width = glue_rep_width + 1

                b_continue2 = True
                while b_continue2:

                    if b_use_fake_seed_row:
                        seed_row_encoding_list, counter_start_val = get_seed_row_encoding_list('B0!', None, iu_tile_set, counter_width, b_final, b_testing=False)
                        #seed_row_encoding_list = ['B0!', '0', '0', '0', '0', '1', '0', '0', '0', '0', '0', '0', '0', 'B1', '#', '1', '0', '0', '0', '0', '0', '0', '0', '0', '1', '1', '1', '#', '1', '0', '0', '0', '0', '0', '0', '0', '1', '0', '0', '0', '0', '#', 'n', '#', 'n', 'B2', '0', '0', '0', '0', '0', '0', '0', 'B3', '$', '#', 'n', '#', 'n', '#', 'n', '#', 'n', '|1', '|2', '#', '1', 'p', '0', '1', '1', '0', '#', '1', 'p', '0', '0', '1', '0', '#', 'n', '#', 'n', 'B4']
                        #seed_row_encoding_list = ['B0!', '0', '0', '0', '1', '0', '0', 'B1', '#', '1', '0', '#', '0', '0', '#', 'n', '#', 'n', 'B2', '0', '0', 'B3', '$', '#', 'n', '#', 'n', '#', 'n', '#', 'n', '|1', '|2', '#', '1', 'p', '0', '1', '#', '1', 'p', '1', '0', '#', 'n', '#', 'n', 'B4']
                        seed_row_encoding_list = ['B0!', '0', '0', '0', '1', '0', '0', 'B1', '#', '1', '0', '#', '0', '0', '#', 'n', '#', 'n', 'B2', '0', '0', 'B3', '$', '#', 'n', '|1', '|2', '#', '1', 'p', 'B4']

                    else:
                        seed_row_encoding_list, counter_start_val = get_seed_row_encoding_list('B0!', quine_builder_tile_set, iu_tile_set, counter_width, b_final)
                    #print(f'seed_row_encoding_list: {seed_row_encoding_list}')

                    # ########################################################
                    # #MJP: Hack to allow simulation!!! 
                    # new_seed_row_encoding_list = []
                    # for sym in seed_row_encoding_list:
                    #     if sym == 'B0!':
                    #         new_seed_row_encoding_list += [sym, '0', '0', '0']
                    #     elif sym == 'B2':
                    #         new_seed_row_encoding_list += [sym, '0', '0']
                    #     elif sym == 'p':
                    #         pass
                    #     else:
                    #         new_seed_row_encoding_list.append(sym)
                    # seed_row_encoding_list = new_seed_row_encoding_list

                    # seed_row_encoding_list = ['B0!', '0', '0', '0', '1', '1', '1', '0', 'B1', '#', '0', '#', 'n', '#', 'n', 'B2', '0', '0', '0', 'B3', '$', '#', 'n', '|1', '|2', '#', '0', '1', '#', 'n', '#', 'n', 'B4']

                    # print(f'seed_row_encoding_list: {seed_row_encoding_list}')
                    # #
                    # ########################################################

                    seed_row_len = len(seed_row_encoding_list)
                    max_needed_counter_val = counter_start_val + seed_row_len - 1
                    max_needed_counter_val_len = len(bin(max_needed_counter_val)[2:])
                    print(f'seed_row_len: {seed_row_len} for counter_width {counter_width}, max_needed_counter_val: {max_needed_counter_val}, max_needed_counter_val_len: {max_needed_counter_val_len}')

                    if counter_width < max_needed_counter_val_len:
                        counter_width += 1
                    else:
                        b_continue2 = False
                        print(f'Done with counter_width {counter_width}')
                        if not b_final:
                            b_continue = True # Loop again to show warnings, etc. 
                            b_final = True
                        if b_final:
                            print(f'Done with final glue_rep_width = {glue_rep_width} and counter_width = {counter_width} (so glues of width {counter_width+1} including strength bit)')

                            (seed_tile_type_name, seed_row_tile_set, seed_row_tile_list) = get_seed_row_tile_set(seed_row_encoding_list)
                            tile_system.tileTypes.update(seed_row_tile_set)
                            tile_system.seedPatch.append((seed_tile_type_name, (0,0)))

                            #draw_tile_set(seed_row_tile_set, output_directory, 'seed_row_tile_set', b_output)

                            tile_system.tileTypes.update(quine_builder_tile_set)

                            # If output directory does not exist, create it!!
                            if not os.path.exists(output_directory):
                                os.makedirs(output_directory)
                            base_file_name = f'{output_directory}/{base_file_name}'
                            
                            tile_system.SaveToFile(base_file_name=base_file_name)

                            get_seed_row_analysis(seed_row_encoding_list, glue_rep_width)

                            seed_row_list_file = f'{base_file_name}_seed_row_list.txt'
                            with open(seed_row_list_file,'w') as fout:
                                fout.write(str(seed_row_encoding_list))
                                print(f"Wrote seed row list to '{seed_row_list_file}'.")

                            print(f'Final tile set saved with {len(tile_system.tileTypes)} tile types. {len(seed_row_tile_set)} for the seed row, {len(quine_builder_tile_set)} for the quine builder, and {len(iu_tile_set)} loaded for the IU tile set.')
                            exit()

        exit()
    
        # num_tile_types = len(tile_set_dict)
        # num_glues_in_dir = get_glue_analysis(tile_set_dict)
        # max_num_glues_in_dir = 0
        # for d in Dirs:
        #     if num_glues_in_dir[d] > max_num_glues_in_dir:
        #         max_num_glues_in_dir = num_glues_in_dir[d]

        # max_num_glues_in_dir =+ len(mod_symbol_value_dict)  # Include the base set of glues
        # bits_needed_for_tile_set_glues = math.ceil(math.log(max_num_glues_in_dir, 2))

        # glue_width = bits_needed_for_tile_set_glues
        # b_keep_going = True
        # while b_keep_going:
        #     tile_encoding_length = 9 + 4 * glue_width
        #     tile_table_length = tile_encoding_length * num_tile_types
        #     bits_needed_for_table_glues = math.ceil(math.log(tile_table_length, 2))

        #     if glue_width >= math.ceil(math.log(max_num_glues_in_dir+tile_table_length, 2)):
        #         b_keep_going = False
        #     else:
        #         glue_width += 1
        #         #print(f'incrementing...')
        
        # print(f'done looping with glue_width = {glue_width}')


